==> ciao-4.6-src-core.tar.gz: ciao-4.6/src/lib/help/AhelpPager.cc <==
// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "AhelpPager.hh"
#include <stdlib.h>
#include <stdio.h>

#define DEF_WIDTH 15

#ifndef __APPLE__
   const char* AhelpPager::DEFAULT_PAGER = "/bin/more";	// default page scroller
#else
   const char* AhelpPager::DEFAULT_PAGER = "/usr/bin/more"; // default page scroller
#endif


// Pager -	Provides a simple class interface to the standard UNIX 
// 		paging programs (less,more,etc).  Instantiate and then
//		use in place of cout,cerr,etc.  Currently has only crude
//		support for formatted output, and could benefit from more
//		generalization.

AhelpPager::Toggle AhelpPager::_toggle = AhelpPager::On;

// Important Note:  disableCtrlC only affects the pager!  It does not
// prevent Ctrl-C from being delivered to the parent process!
AhelpPager::AhelpPager(int disableCtrlC) : _width(0),_precision(6),_showpoint(0),
  _scientific(0),_disableCtrlC(disableCtrlC)
{
	_fp = 0;

  	if (isatty(STDIN_FILENO) && (_toggle==On))
	{
	   const char *pager = getenv("PAGER");

	   if (!pager)
	      pager = DEFAULT_PAGER;

	   if(_disableCtrlC && sighold(SIGINT) == -1)
	     throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

	   if (!(_fp = popen(pager,"w")))
	     throw "Error: Couldn't open pipe to pager";

	   if(_disableCtrlC && sigrelease(SIGINT) == -1) 
	     throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
	}
	else
	   _fp = stdout;
}

// Important Note:  disableCtrlC only affects the pager!  It does not
// prevent Ctrl-C from being delivered to the parent process!
AhelpPager::AhelpPager(char *pager,int disableCtrlC) : _width(0),_precision(6),
  _showpoint(0),_scientific(0),_disableCtrlC(disableCtrlC)
{
	_fp = 0;

	if(_disableCtrlC && sighold(SIGINT) == -1)
	  throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

	if (!pager || !(_fp = popen(pager,"w")))
	  throw "Error: Couldn't open pipe to pager";
	
	if(_disableCtrlC && sigrelease(SIGINT) == -1) 
	  throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
}

AhelpPager::AhelpPager(FILE *fp )
{
	if (fp)
	   _fp = fp;
	else
	   throw "Error: invalid FILE *";
}

AhelpPager::~AhelpPager()
{
  if(_fp != stdout) {
    if(_disableCtrlC && sighold(SIGINT) == -1)
      throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

    if(pclose(_fp) == -1) {
      char err[48];
      sprintf(err, "Error: AhelpPager: Error in pclose, errno: %d", errno);
      throw err;
    }

    if(_disableCtrlC && sigrelease(SIGINT) == -1) 
      throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
  }
}

AhelpPager& AhelpPager::operator<<(const char *str)
{
  if(_width) {
    sprintf(_formatString, "%%%ds", _width);
  } else {
    sprintf(_formatString, "%%s");
  }
    
  	if (str)
           fprintf(_fp, _formatString, str);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const char c)
{
  fprintf(_fp, "%c", c);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const short val)
{
        fprintf(_fp,"%hd",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned short val)
{
        fprintf(_fp,"%d",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const int val)
{
  if(_width) {
    sprintf(_formatString, "%%%dd", _width);
  } else {
    sprintf(_formatString, "%%d");
  }

        fprintf(_fp, _formatString, val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned int val)
{
        fprintf(_fp,"%d",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const long val)
{
        fprintf(_fp,"%ld",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned long val)
{
        fprintf(_fp,"%ld",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const float val)
{
  int oldPrec = _precision;
  int defWidth = DEF_WIDTH;
  
  //
  // Attempts at new rules, hopefully cleaner than what we already have.
  // 28 Jun 2002 pef
  //
  if ( _width && _precision ) {
    short ndigits = 2;
    int pnew = _precision;
    //
    // Use scientific notation if the number is too large...
    // If width is set, make sure there is always a space between columns...
    //
    // For scientific notation:
    // (1) Determine the number of digits in the exponent.
    // (2) Reset _precision to a smaller number if resulting number is too wide.
    //
    //
    if ( (_scientific && val >= 0.0) || 
         (val > 0.0 && log10(val) > (_width-_precision-2)) ) {
      pnew = _width-5-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( (_scientific && val < 0.0) || 
         (val < 0.0 && log10(fabs(val)) > (_width-_precision-3)) ) {
      pnew = _width-6-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( fabs(val) > 0.1 ) {
      if ( _showpoint == 0 ) removeZeroes(val,0,&_precision);
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else if ( val == 0.0 ) {
      if ( _showpoint == 0 ) _precision = 0;
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else {
      int delta = 5;
      if ( val < 0.0 ) delta++;
      pnew = _width-delta-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    }
  } else {
    //
    // Keep old code, for now.
    //
    if (_width)
      defWidth = _width;
    // If user has set _width, but it is not as wide as we need,
    // use scientific notation....
    //
    // Sometimes _width hasn't been set, but we want to use
    // scientific notation anyway.  So, don't make _width
    // a deciding factor in determining whether or not to
    // show in scientific notation.  If _width is zero, then
    // just use a default width when determining if we need
    // scientific notation.  (This avoids printing a number
    // like 10e-9 as "0.00000" when _width is zero).
    //
    // Do continue to use _width, not defWidth, in block
    // below this one, where we show floating point notation --
    // don't want "10" to be shown as "10       ".  SMD 9/27/2001
    if ((((val < 0.0) &&
                 ((log10(fabs(val)) > defWidth-_precision-4) ||
                  (-log10(fabs(val)) > defWidth-_precision-3)))
         || ((val > 0.0) && ((log10(fabs(val)) > defWidth-_precision-3) ||
           (-log10(fabs(val)) > defWidth-_precision-2)))))
      {
        // If val is an exact power of 10, set precision to zero.
        float mantissa = fabs(val)/pow(10, floor(log10(fabs(val))));
        if(floor(mantissa) == mantissa)
          _precision = 0;

        sprintf(_formatString, "%%%d.%de", defWidth, _precision);
      }
    // .... else use standard floating point notation.
    else
      {
        // If val is an integer, use precision of zero.
        if(floor(val) == val)
          _precision = 0;

        if(_width)
          sprintf(_formatString, "%%%d.%df", _width, _precision);
        else
          sprintf(_formatString, "%%.%df", _precision);
      }
  }
  
  fprintf(_fp, _formatString, val);
  _precision = oldPrec;
  return *this;
}

AhelpPager& AhelpPager::operator<<(const double val)
{
  int oldPrec = _precision;
  int defWidth = DEF_WIDTH;
  
  if ( _width && _precision ) {
    short ndigits = 2;
    int pnew = _precision;
    //
    // Use scientific notation if the number is too large...
    // If width is set, make sure there is always a space between columns...
    //
    // For scientific notation:
    // (1) Determine the number of digits in the exponent.
    // (2) Reset _precision to a smaller number if resulting number is too wide.
    //
    //
    if ( (_scientific && val >= 0.0) || 
         (val > 0.0 && log10(val) > (_width-_precision-2)) ) {
      if ( val != 0.0 && log10(fabs(val)) >= 100 ) ndigits++;
      pnew = _width-5-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( (_scientific && val < 0.0) || 
                (val < 0.0 && log10(fabs(val)) > (_width-_precision-3)) ) {
      if ( val != 0.0 && log10(fabs(val)) >= 100 ) ndigits++;
      pnew = _width-6-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( fabs(val) > 0.1 ) {
      if ( _showpoint == 0 ) removeZeroes(val,0,&_precision);
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else if ( val == 0.0 ) {
      if ( _showpoint == 0 ) _precision = 0;
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else {
      if ( log10(fabs(val)) < -99 ) ndigits++;
      int delta = 5;
      if ( val < 0.0 ) delta++;
      pnew = _width-delta-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    }
  } else {
    //
    // Keep old code, for now.
    //
    if (_width)
      defWidth = _width;

    // If user has set _width, but it is not as wide as we need,
    // use scientific notation....
    //
    // Sometimes _width hasn't been set, but we want to use
    // scientific notation anyway.  So, don't make _width
    // a deciding factor in determining whether or not to
    // show in scientific notation.  If _width is zero, then
    // just use a default width when determining if we need
    // scientific notation.  (This avoids printing a number
    // like 10e-9 as " 0.00000" when _width is zero).
    //
    // Do continue to use _width, not defWidth, in block
    // below this one, where we show floating point notation --
    // don't want "10" to be shown as "10       ".  SMD 9/27/2001
    if ((((val < 0.0) &&
                 ((log10(fabs(val)) > defWidth-_precision-4) ||
                  (-log10(fabs(val)) > defWidth-_precision-3)))
       || ((val > 0.0) && ((log10(fabs(val)) > defWidth-_precision-3) ||
           (-log10(fabs(val)) > defWidth-_precision-2)))))
      {
        // If val is an exact power of 10, set precision to zero.
        float mantissa = fabs(val)/pow(10, floor(log10(fabs(val))));
        if(floor(mantissa) == mantissa)
          _precision = 0;

        sprintf(_formatString, "%%%d.%de", defWidth, _precision);
      }
    // .... else use standard (double) floating point notation.
    else
      {
        // If val is an integer, use precision of zero.
        if(floor(val) == val)
          _precision = 0;

        if(_width)
          sprintf(_formatString, "%%%d.%dlf", _width, _precision);
        else
          sprintf(_formatString, "%%.%dlf", _precision);
      }
  }

  fprintf(_fp, _formatString, val);
  _precision = oldPrec;
  return *this;
}

void AhelpPager::removeZeroes(double d,short e,int* p)
{
  double mantissa;
  if ( e == 1 ) {
    mantissa = fabs(d)*pow(10,-floor(log10(fabs(d))));
  } else {
    mantissa = d;
  }
  if ( floor(mantissa) == mantissa ) {
    *p = 0;
  } else {
    short ibreak = 0;
    do {
      if ( *p > 0 ) {
        double x = mantissa*pow(10.0,(*p));
        int y = (int)(floor(x+0.5));
        if ( y % 10 == 0 ) {
          (*p)--;
        } else {
          ibreak = 1;
        }
      } else {
        ibreak = 1;
      }
    } while ( ibreak == 0 );
  }
}

AhelpPager& AhelpPager::operator<<(ostream& (*func)(ostream&))
{
   static ostrstream obuf;
   obuf.clear(); obuf.seekp(ios::beg);
   obuf << *func << ends;
   fprintf(_fp,"%s",obuf.str());
   fflush(_fp);
   return *this;
}

// Signal handling will be made its own function in libascgen.so,
// when we figure out the proper source code location.
int AhelpPager::sighold(int signal)
{
  sigset_t signals;
  sigemptyset(&signals);
  sigaddset(&signals, signal);
  return sigprocmask(SIG_BLOCK, &signals, 0);
}

// Signal handling will be made its own function in libascgen.so,
// when we figure out the proper source code location.
int AhelpPager::sigrelease(int signal)
{
  sigset_t signals;
  sigemptyset(&signals);
  sigaddset(&signals, signal);
  return sigprocmask(SIG_UNBLOCK, &signals, 0);
}

==> ciao-4.8-src-core.tar.gz: ciao-4.8/src/lib/help/AhelpPager.cc <==
// 
//  Copyright (C) 2007,2015  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "AhelpPager.hh"
#include <stdlib.h>
#include <stdio.h>

#define DEF_WIDTH 15

#ifndef __APPLE__
   const char* AhelpPager::DEFAULT_PAGER = "/bin/more";	// default page scroller
#else
   const char* AhelpPager::DEFAULT_PAGER = "/usr/bin/more"; // default page scroller
#endif


// Pager -	Provides a simple class interface to the standard UNIX 
// 		paging programs (less,more,etc).  Instantiate and then
//		use in place of cout,cerr,etc.  Currently has only crude
//		support for formatted output, and could benefit from more
//		generalization.

AhelpPager::Toggle AhelpPager::_toggle = AhelpPager::On;

// Important Note:  disableCtrlC only affects the pager!  It does not
// prevent Ctrl-C from being delivered to the parent process!
AhelpPager::AhelpPager(int disableCtrlC) : _width(0),_precision(6),_showpoint(0),
  _scientific(0),_disableCtrlC(disableCtrlC)
{
	_fp = 0;

  	if (isatty(STDIN_FILENO) && (_toggle==On))
	{
	   const char *pager = getenv("PAGER");

	   if (!pager)
	      pager = DEFAULT_PAGER;

	   if(_disableCtrlC && sighold(SIGINT) == -1)
	     throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

	   if (!(_fp = popen(pager,"w")))
	     throw "Error: Couldn't open pipe to pager";

	   if(_disableCtrlC && sigrelease(SIGINT) == -1) 
	     throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
	}
	else
	   _fp = stdout;
}

// Important Note:  disableCtrlC only affects the pager!  It does not
// prevent Ctrl-C from being delivered to the parent process!
AhelpPager::AhelpPager(char *pager,int disableCtrlC) : _width(0),_precision(6),
  _showpoint(0),_scientific(0),_disableCtrlC(disableCtrlC)
{
	_fp = 0;

	if(_disableCtrlC && sighold(SIGINT) == -1)
	  throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

	if (!pager || !(_fp = popen(pager,"w")))
	  throw "Error: Couldn't open pipe to pager";
	
	if(_disableCtrlC && sigrelease(SIGINT) == -1) 
	  throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
}

AhelpPager::AhelpPager(FILE *fp )
{
	if (fp)
	   _fp = fp;
	else
	   throw "Error: invalid FILE *";
}

AhelpPager::~AhelpPager()
{
  if(_fp != stdout) {
    if(_disableCtrlC && sighold(SIGINT) == -1)
      throw "Error: AhelpPager: Couldn't turn off Ctrl-C";

    if(pclose(_fp) == -1) {
      char err[48];
      sprintf(err, "Error: AhelpPager: Error in pclose, errno: %d", errno);
      throw err;
    }

    if(_disableCtrlC && sigrelease(SIGINT) == -1) 
      throw "Error: AhelpPager: Couldn't turn Ctrl-C back on";
  }
}

AhelpPager& AhelpPager::operator<<(const char *str)
{
  if(_width) {
    sprintf(_formatString, "%%%ds", _width);
  } else {
    sprintf(_formatString, "%%s");
  }
    
  	if (str)
           fprintf(_fp, _formatString, str);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const char c)
{
  fprintf(_fp, "%c", c);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const short val)
{
        fprintf(_fp,"%hd",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned short val)
{
        fprintf(_fp,"%d",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const int val)
{
  if(_width) {
    sprintf(_formatString, "%%%dd", _width);
  } else {
    sprintf(_formatString, "%%d");
  }

        fprintf(_fp, _formatString, val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned int val)
{
        fprintf(_fp,"%d",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const long val)
{
        fprintf(_fp,"%ld",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const unsigned long val)
{
        fprintf(_fp,"%ld",val);
	return *this;
}

AhelpPager& AhelpPager::operator<<(const float val)
{
  int oldPrec = _precision;
  int defWidth = DEF_WIDTH;
  
  //
  // Attempts at new rules, hopefully cleaner than what we already have.
  // 28 Jun 2002 pef
  //
  if ( _width && _precision ) {
    short ndigits = 2;
    int pnew = _precision;
    //
    // Use scientific notation if the number is too large...
    // If width is set, make sure there is always a space between columns...
    //
    // For scientific notation:
    // (1) Determine the number of digits in the exponent.
    // (2) Reset _precision to a smaller number if resulting number is too wide.
    //
    //
    if ( (_scientific && val >= 0.0) || 
         (val > 0.0 && log10(val) > (_width-_precision-2)) ) {
      pnew = _width-5-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( (_scientific && val < 0.0) || 
         (val < 0.0 && log10(fabs(val)) > (_width-_precision-3)) ) {
      pnew = _width-6-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( fabs(val) > 0.1 ) {
      if ( _showpoint == 0 ) removeZeroes(val,0,&_precision);
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else if ( val == 0.0 ) {
      if ( _showpoint == 0 ) _precision = 0;
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else {
      int delta = 5;
      if ( val < 0.0 ) delta++;
      pnew = _width-delta-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    }
  } else {
    //
    // Keep old code, for now.
    //
    if (_width)
      defWidth = _width;
    // If user has set _width, but it is not as wide as we need,
    // use scientific notation....
    //
    // Sometimes _width hasn't been set, but we want to use
    // scientific notation anyway.  So, don't make _width
    // a deciding factor in determining whether or not to
    // show in scientific notation.  If _width is zero, then
    // just use a default width when determining if we need
    // scientific notation.  (This avoids printing a number
    // like 10e-9 as "0.00000" when _width is zero).
    //
    // Do continue to use _width, not defWidth, in block
    // below this one, where we show floating point notation --
    // don't want "10" to be shown as "10       ".  SMD 9/27/2001
    if ((((val < 0.0) &&
                 ((log10(fabs(val)) > defWidth-_precision-4) ||
                  (-log10(fabs(val)) > defWidth-_precision-3)))
         || ((val > 0.0) && ((log10(fabs(val)) > defWidth-_precision-3) ||
           (-log10(fabs(val)) > defWidth-_precision-2)))))
      {
        // If val is an exact power of 10, set precision to zero.
        float mantissa = fabs(val)/pow(10, floor(log10(fabs(val))));
        if(floor(mantissa) == mantissa)
          _precision = 0;

        sprintf(_formatString, "%%%d.%de", defWidth, _precision);
      }
    // .... else use standard floating point notation.
    else
      {
        // If val is an integer, use precision of zero.
        if(floor(val) == val)
          _precision = 0;

        if(_width)
          sprintf(_formatString, "%%%d.%df", _width, _precision);
        else
          sprintf(_formatString, "%%.%df", _precision);
      }
  }
  
  fprintf(_fp, _formatString, val);
  _precision = oldPrec;
  return *this;
}

AhelpPager& AhelpPager::operator<<(const double val)
{
  int oldPrec = _precision;
  int defWidth = DEF_WIDTH;
  
  if ( _width && _precision ) {
    short ndigits = 2;
    int pnew = _precision;
    //
    // Use scientific notation if the number is too large...
    // If width is set, make sure there is always a space between columns...
    //
    // For scientific notation:
    // (1) Determine the number of digits in the exponent.
    // (2) Reset _precision to a smaller number if resulting number is too wide.
    //
    //
    if ( (_scientific && val >= 0.0) || 
         (val > 0.0 && log10(val) > (_width-_precision-2)) ) {
      if ( val != 0.0 && log10(fabs(val)) >= 100 ) ndigits++;
      pnew = _width-5-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( (_scientific && val < 0.0) || 
                (val < 0.0 && log10(fabs(val)) > (_width-_precision-3)) ) {
      if ( val != 0.0 && log10(fabs(val)) >= 100 ) ndigits++;
      pnew = _width-6-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    } else if ( fabs(val) > 0.1 ) {
      if ( _showpoint == 0 ) removeZeroes(val,0,&_precision);
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else if ( val == 0.0 ) {
      if ( _showpoint == 0 ) _precision = 0;
      sprintf(_formatString,"%%%d.%df",_width,_precision);
    } else {
      if ( log10(fabs(val)) < -99 ) ndigits++;
      int delta = 5;
      if ( val < 0.0 ) delta++;
      pnew = _width-delta-ndigits;
      if ( pnew >= 0 && pnew < _precision ) _precision = pnew;
      if ( _showpoint == 0 ) removeZeroes(val,1,&_precision);
      sprintf(_formatString,"%%%d.%de",_width,_precision);
    }
  } else {
    //
    // Keep old code, for now.
    //
    if (_width)
      defWidth = _width;

    // If user has set _width, but it is not as wide as we need,
    // use scientific notation....
    //
    // Sometimes _width hasn't been set, but we want to use
    // scientific notation anyway.  So, don't make _width
    // a deciding factor in determining whether or not to
    // show in scientific notation.  If _width is zero, then
    // just use a default width when determining if we need
    // scientific notation.  (This avoids printing a number
    // like 10e-9 as " 0.00000" when _width is zero).
    //
    // Do continue to use _width, not defWidth, in block
    // below this one, where we show floating point notation --
    // don't want "10" to be shown as "10       ".  SMD 9/27/2001
    if ((((val < 0.0) &&
                 ((log10(fabs(val)) > defWidth-_precision-4) ||
                  (-log10(fabs(val)) > defWidth-_precision-3)))
       || ((val > 0.0) && ((log10(fabs(val)) > defWidth-_precision-3) ||
           (-log10(fabs(val)) > defWidth-_precision-2)))))
      {
        // If val is an exact power of 10, set precision to zero.
        float mantissa = fabs(val)/pow(10, floor(log10(fabs(val))));
        if(floor(mantissa) == mantissa)
          _precision = 0;

        sprintf(_formatString, "%%%d.%de", defWidth, _precision);
      }
    // .... else use standard (double) floating point notation.
    else
      {
        // If val is an integer, use precision of zero.
        if(floor(val) == val)
          _precision = 0;

        if(_width)
          sprintf(_formatString, "%%%d.%dlf", _width, _precision);
        else
          sprintf(_formatString, "%%.%dlf", _precision);
      }
  }

  fprintf(_fp, _formatString, val);
  _precision = oldPrec;
  return *this;
}

void AhelpPager::removeZeroes(double d,short e,int* p)
{
  double mantissa;
  if ( e == 1 ) {
    mantissa = fabs(d)*pow(10,-floor(log10(fabs(d))));
  } else {
    mantissa = d;
  }
  if ( floor(mantissa) == mantissa ) {
    *p = 0;
  } else {
    short ibreak = 0;
    do {
      if ( *p > 0 ) {
        double x = mantissa*pow(10.0,(*p));
        int y = (int)(floor(x+0.5));
        if ( y % 10 == 0 ) {
          (*p)--;
        } else {
          ibreak = 1;
        }
      } else {
        ibreak = 1;
      }
    } while ( ibreak == 0 );
  }
}

AhelpPager& AhelpPager::operator<<(ostream& (*func)(ostream&))
{
   ostringstream obuf;
   obuf << *func;
   fprintf(_fp,"%s", obuf.str().c_str());
   fflush(_fp);
   return *this;
}

// Signal handling will be made its own function in libascgen.so,
// when we figure out the proper source code location.
int AhelpPager::sighold(int signal)
{
  sigset_t signals;
  sigemptyset(&signals);
  sigaddset(&signals, signal);
  return sigprocmask(SIG_BLOCK, &signals, 0);
}

// Signal handling will be made its own function in libascgen.so,
// when we figure out the proper source code location.
int AhelpPager::sigrelease(int signal)
{
  sigset_t signals;
  sigemptyset(&signals);
  sigaddset(&signals, signal);
  return sigprocmask(SIG_UNBLOCK, &signals, 0);
}
